"""
The logic for saving an image
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111

import os

from asynciojobs import Scheduler, Job

from rhubarbe.collector import Collector
from rhubarbe.leases import Leases
from rhubarbe.config import Config


class ImageSaver:

    def __init__(self, node, image, radical,            # pylint: disable=r0913
                 message_bus, display, comment):
        self.node = node
        self.image = image
        self.radical = radical
        self.message_bus = message_bus
        self.display = display
        self.comment = comment
        #
        self.collector = None


    async def feedback(self, field, msg):
        await self.message_bus.put({field: msg})


    # this is exactly as imageloader
    async def stage1(self):
        the_config = Config()
        idle = int(the_config.value('nodes', 'idle_after_reset'))
        await self.node.reboot_on_frisbee(idle)


    # this is synchroneous
    def nextboot_cleanup(self):
        """
        Remove nextboot symlinks for all nodes in this selection
        so next boot will be off the harddrive
        """
        self.node.manage_nextboot_symlink('harddrive')


    async def start_collector(self):
        self.collector = Collector(self.image, self.message_bus)
        port = await self.collector.start()
        return port


    async def stage2(self, reset):
        """
        run collector (a netcat server)
        then wait for the node to be telnet-friendly,
        then run imagezip on the node
        reset node when finished unless reset is False
        """
        # start_frisbeed will return the ip+port to use
        await self.feedback('info', f"Saving image from {self.node}")
        port = await self.start_collector()
        result = await self.node.run_imagezip(port, reset,
                                              self.radical, self.comment)
        # we can now kill the server
        self.collector.stop_nowait()
        if not result:
            await self.feedback('info', "Failed to save disk image")
        return result


    async def run(self, reset):
        leases = Leases(self.message_bus)
        await self.feedback('authorization', 'checking for a valid lease')
        valid = await leases.booked_now_by_me()
        if not valid:
            await self.feedback('authorization',
                                "Access refused : you have no lease"
                                " on the testbed at this time")
            return False
        await (self.stage1()
               if reset
               else self.feedback('info', "Skipping stage1"))
        return await self.stage2(reset)


    def mark_image_as_partial(self):
        # never mind if that fails, we might call this before
        # the file is created
        try:
            os.rename(self.image, self.image + ".partial")
        except Exception:                               # pylint: disable=w0703
            pass


    def cleanup(self):
        if self.collector:
            self.collector.stop_nowait()
        self.nextboot_cleanup()
        self.display.epilogue()


    def main(self, reset, timeout):
        mainjob = Job(self.run(reset), critical=True)
        displayjob = Job(self.display.run(), forever=True, critical=True)

        scheduler = Scheduler(mainjob, displayjob,
                              timeout=timeout,
                              critical=False)

        try:
            is_ok = scheduler.run()
            if not is_ok:
                scheduler.debrief()
                self.display.set_goodbye(
                    f"rhubarbe-save failed: {scheduler.why()}")
                return 1
            return 0 if mainjob.result() else 1
        except KeyboardInterrupt:
            self.display.set_goodbye("rhubarbe-save : keyboard interrupt, bye")
            return 1
        finally:
            self.cleanup()



"""
Parse inventory json about nodes
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, r1705

import json

from rhubarbe.singleton import Singleton
from rhubarbe.config import Config


class Inventory(metaclass=Singleton):

    def __init__(self):
        the_config = Config()
        with open(the_config.value('testbed', 'inventory_nodes_path')) as feed:
            self._nodes = json.load(feed)

    def _locate_entry_from_key(self, key, value):
        """
        search for an entry that has given hostname
        returns a tuple (host, key)
        e.g.
        _locate_entry_from_key('hostname', 'reboot01') =>
         ( { 'cmc' : {...}, 'control' : {...}, 'data' : {...} }, 'cmc' )
         """
        for host in self._nodes:
            for k, v in host.items():                   # pylint: disable=c0103
                if v[key] == value:
                    return host, k
        return None, None

    def attached_hostname_info(self, hostname,
                               interface_key='control', info_key='hostname'):
        """
        locate the entry that has at least one hostname equal to 'hostname'
        and returns the 'hostname' attached to that key
        e.g.
        attached_hostname('reboot01', 'control') => 'fit01'
        """
        host, _ = self._locate_entry_from_key('hostname', hostname)
        if host and interface_key in host:
            return host[interface_key][info_key]
        return None

    def control_ip_from_any_ip(self, ipaddr):
        host, _ = self._locate_entry_from_key('ip', ipaddr)
        if host:
            return host['control']['ip']
        return None

    def display(self, verbose=False):
        def cell_repr(key, value, verbose):
            if not verbose:
                return f"{key}:{value['hostname']}"
            else:
                return f"{key}:{value['hostname']}[{value['mac']}]"
        print(20*'-', "INVENTORY CONTENTS")
        for node in self._nodes:
            print(" ".join([cell_repr(k, v, verbose)
                            for k, v in node.items()]))
        print(20*'-', "INVENTORY END")

    def one_control_interface(self):
        return self._nodes[0]['control']['ip']

    def all_control_hostnames(self):
        return (node['control']['hostname'] for node in self._nodes)



"""
The contents of one directory that holds images
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, c0103

import sys
import os
import time
import re
#import os.path
# should go away too
import glob
from pathlib import Path

# to indicate that 0 is OK and others are KO
from typing import List
OsRetcod = int

from rhubarbe.config import Config
from rhubarbe.singleton import Singleton

debug = False

# use __ instead of == because = ruins bash completion
saving_keyword = 'saving'
saving_sep = '__'
saving_time_format = "%Y-%m-%d@%H-%M"

sep_re_pattern = "[-_=]{2}"
date_re_pattern = "[0-9]{4}-[0-9]{2}-[0-9]{2}@[0-9]{2}.[0-9]{2}"
# this will be fed into .format() - needs double {{ and }}
hostname_re_pattern_f = "{regularname}[0-9]{{2}}"

sep_glob_pattern = "[-_=]" * 2
date_glob_pattern = "[0-9-@]" * 16
hostname_glob_pattern_f = "{regularname}[0-9][0-9]"


class ImagesRepo(metaclass=Singleton):
    def __init__(self):
        the_config = Config()
        self.repo = Path(the_config.value('frisbee', 'images_dir'))
        self.default_name = the_config.value('frisbee', 'default_image')
        # for building hostname patterns
        self.regularname = the_config.value('testbed', 'regularname')

    suffix = ".ndz"

    def default(self) -> str:
        return str(self.repo / self.default_name)

    def where_to_save(self, nodename, radical) -> str:
        """
        given a nodename, plus a user-provided image name
        computes the actual path where to store an image
        as of 3.0.7, root is deemed to not be likely to build images
        so no more special case for uid==0
        """
        parts = [saving_keyword, nodename,
                 time.strftime(saving_time_format), radical]
        base = saving_sep.join(parts) + self.suffix

        # regular users store image where they are
        # a more sophisticated approach would be to split radical
        # to find about another directory, but well..
        return base

    def radical_part(self, path):
        """
        incoming can be a filename, possibly without an extension
        the saving__ blabla is extracted and only the part
        given at save-time is returned
        it is kind of the reverse of where_to_save
        """
        incoming = path.name.replace(self.suffix, "")
        parts = [
            saving_keyword,
            hostname_glob_pattern_f.format(regularname=self.regularname),
            date_glob_pattern,
            "(?P<radical>.+)"]
        glob_pattern = re.compile(saving_sep.join(parts))
        match = glob_pattern.match(incoming)
        if match:
            return match.group('radical')
        return incoming

    class ImageInfo:                              # pylint: disable=r0902,r0903
        def __init__(self, repo, filename=None):
            if filename is not None:
                # a valid path
                self.repo = repo
                self.filename = filename
            else:
                self.repo = repo.parent
                self.filename = repo.name
            self.fill_info()

        def fill_info(self):
            try:
                with open(self.filename) as fake:
                    fake.read(1)
                self.readable = True
            except:
                self.readable = False
                return
            self.radical = self.repo.radical_part(self.filename)
            stat = os.stat(self.filename)
            self.mtime = stat.st_mtime
            self.size = stat.st_size
            self.basename = os.path.splitext(os.path.basename(self.filename))
            self.is_official = self.radical == self.basename
            self.is_latest = None

        def __str__(self):
            return "{}{}{}{} {}".format(
                "*" if self.is_latest else " ",
                "O" if self.is_official else " ",
                "!" if not self.readable else " ",
                self.filename,
                ImagesRepo.bytes2human(self.size))

    def _locate_infos(self, input_name, look_in_global):
        """
        search an image from its name
        * input_name: a filename, possibly without extension,
          possibly without saving_blabla
        * look_in_global: do we search in the global repo or just in .

        returns a list of ImageInfo objects
        (filename, radical, is_default, is_latest)
        where
        * filename is a valid filename that can be open()
        * radical
        is_default is a boolean that says if this is the

        If the image name actually denotes a filename (i.e. with .ndz) then
        no completion takes place

        Otherwise, all 'saving' images are also considered
        and the latest one is picked
        """

        # when look_in_global is set, but we are already in the repo:
        # turn it off
        if (look_in_global and
            Path(".").absolute() == self.repo.absolute()):
            look_in_global = False

        # existing filename ?
        candidate = input_name
        local_info = self.ImageInfo(Path('.'), input_name)
        if local_info.readable:
            return [local_info]
        if look_in_global and not Path(input_name).is_absolute():
            global_info = self.ImageInfo(self.repo, input_name)
            if global_info.readable:
                return [global_info]

        # it's not a plain filename, so explore more options
        glob_pattern = (
            saving_sep.join([
                saving_keyword,
                hostname_glob_pattern_f.format(regularname=self.regularname),
                date_glob_pattern,
                input_name
            ]))
        globs = [Path(".").glob(input_name + self.suffix),
                 Path(".").glob(glob_pattern + self.suffix)]
        if debug:
            print("glob_pattern", glob_pattern)
        if look_in_global and not Path(input_name).is_absolute():
            globs += [self.repo.glob(input_name + self.suffix),
                      self.repo.glob(glob_pattern + self.suffix)]
        # run these patterns
        found = []
        from itertools import chain
        for matched in chained(*globs):
            info = self.ImageInfo(matched)
            if info.readable:
                found.append(info)
        return found

    def _locate_infos_sorted(self, input_name, look_in_global=True):
        details = self._locate_infos(input_name, look_in_global)
        details.sort(key=lambda info: info.mtime, reverse=True)
        if details:
            details[0].is_latest = True                 # pylint: disable=w0201
        return details

    @staticmethod
    def is_root():
        # surprisingly, even when running under sudo we have all 3 set to 0
        real, effective, saved = os.getresuid()         # pylint: disable=e1101
        is_root = (real == 0) and (effective == 0) and (saved == 0)
        return is_root

    def locate_image(self, image_name, look_in_global=None) -> str:
        if look_in_global is None:
            look_in_global = self.is_root()
        infos = self._locate_infos_sorted(image_name, look_in_global)
        return infos[0].filename if infos else []

    @staticmethod
    def bytes2human(n, repr_format="{value:.3f} {symbol}"):
        """
        >>> bytes2human(10000)
        '9K'
        >>> bytes2human(100001221)
        '95M'
        """
        symbols = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
        prefix = {}
        for i, s in enumerate(symbols[1:]):
            prefix[s] = 1 << (i+1)*10
        for symbol in reversed(symbols[1:]):
            if n >= prefix[symbol]:
                value = float(n) / prefix[symbol]
                return repr_format.format(value=value, symbol=symbol)
        return repr_format.format(symbol=symbols[0], value=n)

    # entry point for rhubarbe images
    def images(self, focus, verbose=False,
               sort_by='date', reverse=False,
               *, show_local, show_global) -> OsRetcod:
        return self.display(focus, verbose=verbose, sort_by=sort_by,
                            reverse=reverse, human_readable=True,
                            show_local=show_local, show_global=show_global)

    def display(self, focus, verbose,        # pylint:disable=r0912,r0913,r0914
                sort_by, reverse, human_readable,
                *, show_local, show_global) -> OsRetcod:
        # temporarily
        return self._gather(focus, verbose, sort_by, reverse, human_readable,
                            path=None)

    def _gather(self, focus, verbose,        # pylint:disable=r0912,r0913,r0914
                sort_by, reverse, human_readable,
                *, path) -> List[ImageInfo]:
        # show available images in some sensible way
        #
        # (*) focus: a list of re patterns
        # (.) if empty, everything is displayed
        # (.) otherwise only files that match one of these names
        #        are displayed (together with their symlinks)
        # (*) sort_by, reverse: how to sort
        # (*) human_readable: bool, show raw size or use units like MB or Gb
        #
        # We want to show all the names attached to a given image
        # through symlinks;  this is why all the filenames are
        # grouped by clusters; there is one cluster per real image file
        #
        # so what we do is
        # (1) scan all files and gather them by clusters
        #     that point at the same file
        #     for this we use a (inode, mtime) tuple, same value = same file
        # (2) for each cluster, we distinguish between
        #     'single' clusters, i.e. an iamge file without any symlink
        #     'annotated' clusters have at least one symlink
        #
        # default behaviour is to show only annotated images
        # when verbose is turned on, all clusters are shown

        print("==================== Images repository {}".format(self.repo))
        if not verbose:
            print("==== use with --verbose to see images without a symlink")
        if focus:
            print("==== images matching any of", " ".join(focus))

        # gather one info per file, with
        # a key (inode, mtime, size) (same key = same file),
        # prefix (filename, no extension),
        # isalias
        infos = []
        for path in glob.glob("{}/*.ndz".format(self.repo)):
            prefix = os.path.basename(path).replace(".ndz", "")
            try:
                stat = os.stat(path)
            except Exception:                           # pylint: disable=w0703
                print("WARNING - Cannot stat {} (dangling symlink ?) - ignored"
                      .format(path))
                continue
            stat_tuple = (stat.st_size, stat.st_mtime, stat.st_ino, )
            info = {'stat_tuple': stat_tuple, 'prefix': prefix,
                    'isalias': os.path.islink(path)}
            infos.append(info)
        # gather by same file (same stat_tuple)
        clusters = {}
        for info in infos:
            stat_tuple = info['stat_tuple']
            clusters.setdefault(stat_tuple, [])
            clusters[stat_tuple].append(info)

        # sort infos in one cluster, so that real files show up first
        def sort_info(info):
            return info['isalias']
        for cluster in clusters.values():
            cluster.sort(key=sort_info)

        # sort clusters
        cluster_items = list(clusters.items())

        def sort_clusters(k_v):
            # k_v is a tuple with a key and a list of infos
            # k_v[0] is a key
            return k_v[0][0] if sort_by == 'size' \
                else k_v[0][1]
        cluster_items.sort(key=sort_clusters, reverse=reverse)

        # is a given cluster filtered by focus
        def in_focus(cluster, focus):
            # an empty focus means to show all
            if not focus:
                return True
            # in this context a cluster is just a list of infos
            for info in cluster:
                for filtr in focus:
                    # rough : no regexp, just find the filter or not
                    if info['prefix'].find(filtr) >= 0:
                        return True
            return False

        for stat_tuple, cluster in cluster_items:
            # skip the cluster if not in focus
            if not in_focus(cluster, focus):
                continue
            # without the verbose flag, show only clusters
            # that have at least one symlink
            if not verbose:
                # do we have at least one symlink ?
                if not any([info['isalias'] for info in cluster]):
                    # ignoring single clusters in non-verbose mode
                    continue
            shown = False
            for info in cluster:
                (size, mtime, _) = info['stat_tuple']
                print_size = size if not human_readable \
                    else ImagesRepo.bytes2human(size)
                date = time.strftime("%Y-%m-%d %H:%M", time.localtime(mtime))
                prefix = info['prefix']
                if not shown:
                    print("{:<18}{:>12}B {:40}"
                          .format(date, print_size, prefix))
                    shown = True
                else:
                    print("{:>31} {:40}".format("a.k.a.", prefix))
        return 0

    # rhubarbe resolve
    def resolve(self, focus, verbose, reverse) -> OsRetcod:
        for incoming in focus:
            infos = self._locate_infos_sorted(incoming)
            if not infos:
                print("Image {} not found".format(incoming), file=sys.stderr)
            else:
                if not verbose:
                    # main purpose is to print out the right name
                    for info in infos:
                        if info.is_latest:
                            print(info.filename)
                else:
                    print(8*'*', "image {}".format(incoming))
                    # it comes in the wrong order, latest first
                    if not reverse:
                        infos.reverse()
                    for info in infos:
                        print(info)

    def share(self, images, alias,           # pylint:disable=r0912,r0913,r0914
              dry_run, force, clean) -> OsRetcod:
        """
        A utility to install one or several images in the shared repo

        Parameters:
          images: a list of image names (not searched in global repo)
            to transfer
          alias: a name for the installed image -
            only if a single image is provided

        Each provided image gets renamed into the radical of origin name,
        i.e. with the extra saving__*blabla removed

        (.) When a single image is given, and 'alias' is set,
            then a symlink from alias to radical is created

        (.) identity
        root is allowed to override destination, not regular users

        # return 0 if all goes well, 1 otherwise
        """

        # check args
        if len(images) > 1 and alias:
            print("alias can be specified only with one image")
            return 1

        is_root = self.is_root()
        # compute dry_run if not set on the command line
        if dry_run is None:
            dry_run = False if is_root else True

        # first pass
        # compute a list of tuples (origin, destination)
        for image in images:
            moves = []
            removes = []   # list of tuples
            symlinks = []  # list of tuples
            origin = self.locate_image(image, look_in_global=is_root)
            if not origin:
                print("Could not find image {} - ignored".format(image))
                continue
            matching_infos = \
                self._locate_infos_sorted(self.radical_part(origin),
                                          look_in_global=is_root)
            # at this point there's at most one that is marked official
            # but if the user has given us a full path, that's the one
            # that is going to be official, so must not be removed either
            for info in matching_infos:
                if info.filename == origin:
                    info.is_official = True              # pylint:disable=w0201

            radical = self.radical_part(origin)
            if not radical:
                print("WARNING: Could not guess radical part in {}\n"
                      "  you will need to give one with -d".format(origin))
                continue
            destination = os.path.join(self.repo, radical + ".ndz")
            if os.path.exists(destination) and not force:
                print("WARNING: Destination {} already exists - ignored"
                      .format(destination))
            else:
                moves.append((origin, destination))  # append a tuple
            if alias:
                symlinks.append((radical + ".ndz",   # ditto
                                 os.path.join(self.repo, alias + ".ndz")))
            else:
                print("Warning : share without an alias")
            if clean:
                # item # 0 is the one selected for being moved
                for info in matching_infos[1:]:
                    # among the other ones, DON't remove the official
                    # since it's useless - or harmful
                    if not info.is_official:
                        removes.append(info.filename)

            print(8*'*', image)
            matching_infos.reverse()
            for info in matching_infos:
                print(info)

            def do_dry_run(*args):
                print("DRY-RUN: would do")
                print(*args)

            for remove in removes:
                if dry_run:
                    do_dry_run("rm {}".format(remove))
                else:
                    print("Removing {}".format(remove))
                    os.unlink(remove)

            for origin, destination in moves:
                if dry_run:
                    do_dry_run("mv {} {}".format(origin, destination))
                else:
                    print("Moving {} -> {}".format(origin, destination))
                    os.rename(origin, destination)

            for origin, destination in symlinks:
                if dry_run:
                    do_dry_run("ln -sf {} {}".format(origin, destination))
                else:
                    print("Creating symlink {} -> {}".
                          format(destination, origin))
                    if os.path.exists(destination):
                        os.unlink(destination)
                    os.symlink(origin, destination)

        return 0



"""
The PlcApiProxy class allows to create an authenticated xmlrpc
connection to a myplc server; typically r2labapi.inria.fr
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import getpass

import ssl

# from aioxmlrpc.client import ServerProxy
from xmlrpc.client import ServerProxy


class PlcApiProxy(ServerProxy):                         # pylint: disable=r0903

    # use the standard plcapi scheme to run on /PLCAPI/
    # always use password auth in this first rough version
    # could be improved by using session authentication
    def __init__(self, url, email=None, password=None, debug=False):
        self.url = url
        self.email = email
        self.password = password
        self.debug = debug
        ###
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        context.check_hostname = False
        ServerProxy.__init__(
            self, self.url, allow_none=True, context=context
        )

    def __auth__(self, anonymous):
        if anonymous:
            return {'AuthMethod': 'anonymous'}
        else:
            if not self.email:
                self.email = input("Enter plcapi email (login) : ")
            if not self.password:
                self.password = getpass.getpass(
                    f"Enter plcapi password for {self.email} : ")
            return {'AuthMethod': 'password',
                    'Username': self.email,
                    'AuthString': self.password}

    def __getattr__(self, attr):
        """
        pass the authentication along for all calls
        """
        # the default is to use authenticated calls
        # because this is the majority of the plcapi calls
        def fun(*args, anonymous=False, **kwds):
            if self.debug:
                auth_msg = "[auth]" if not anonymous else "[anon]"
                print(f"-> Sending {auth_msg} {attr} on {self} "
                      f"with args={args} and kwds={kwds}")
            actual_fun = ServerProxy.__getattr__(
                self, attr)
            try:
                retcod = actual_fun(self.__auth__(anonymous), *args, **kwds)
                if self.debug:
                    print(f"<- Received {retcod}")
                return retcod
            except Exception as exc:
                print(f"ignored exception in {attr} : {exc}")
        return fun

    def __str__(self):
        return f"PLCAPIproxy@{self.url}"



#!/usr/bin/env python3

"""
Shared commodities for conveniently selecting a set of nodes.
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111

import os

from rhubarbe.config import Config


class MisformedRange(Exception):

    def __init__(self, rangetext):
        self.rangetext = rangetext
        super().__init__(self)

    def __str__(self):
        return f"Misformed node range '{self.rangetext}'"


class Selector:

    # typically regularname='fit' and rebootname='reboot'
    # so that fit01 and reboot01 are names that resolve
    def __init__(self):
        the_config = Config()
        self.regularname = the_config.value('testbed', 'regularname')
        self.rebootname = the_config.value('testbed', 'rebootname')
        self.set = set()

    def __repr__(self):
        return "<Selector " + " ".join(self.node_names()) + ">"

    def __str__(self):
        return "Selected nodes: " + " ".join(self.node_names())

    def add_or_delete(self, index, add_if_true):
        if add_if_true:
            self.set.add(index)
        # set.remove triggers KeyError
        elif index in self.set:
            self.set.remove(index)

    # range is a shell arg, like fit01, fit1, 1, 1-12, ~25
    def add_range(self, range_spec):
        range_spec = (range_spec
                      .replace(self.regularname, "")
                      .replace(self.rebootname, ""))
        commas = range_spec.split(',')
        for comma in commas:
            adding = True
            if comma.startswith('~'):
                adding = False
                comma = comma[1:]
            try:
                items = [int(x) for x in comma.split('-')]
            except Exception:
                # safer to exit abruptly;
                # common mistake is to forget -i, like in
                # rload image-radical 1 2 3
                # if we just ignore this situation, the wrong sentence
                # leads to a totally different behaviour
                raise MisformedRange(comma)
            if len(items) >= 4:
                print(f"Ignored arg {comma}")
                continue
            elif len(items) == 3:
                lower, upper, step = items
                for i in range(lower, upper+1, step):
                    self.add_or_delete(i, adding)
            elif len(items) == 2:
                lower, upper = items
                for i in range(lower, upper+1):
                    self.add_or_delete(i, adding)
            else:
                i = items[0]
                self.add_or_delete(i, adding)

    # generators
    def node_names(self):
        return (f"{self.regularname}{i:02}" for i in sorted(self.set))

    def cmc_names(self):
        return (f"{self.rebootname}{i:02}" for i in sorted(self.set))

    def __len__(self):
        return len(self.set)

    def is_empty(self):
        return not self.set

    def use_all_scope(self):
        the_config = Config()
        self.add_range(the_config.value('testbed', 'all_scope'))

####################
# convenience tools shared by all commands that need this sort of selection
# maybe we should have specialized ArgumentParser instead


def add_selector_arguments(arg_parser):
    arg_parser.add_argument(
        "-a", "--all-nodes", action='store_true', default=False,
        help="""
        add the contents of the testbed.all_scope config variable in the mix
        this can be combined with ranges, like e.g.
        -a ~4-16-2
        """)
    arg_parser.add_argument(
        "ranges", nargs="*",
        help="""
        nodes can be specified one by one, like: 1 004 fit01 reboot01;
        or by range (inclusive) like: 2-12, fit4-reboot12;
        or by range+step like:
        4-16-2 which would be all even numbers from 4 to 16;
        ranges can also be excluded with '~', so ~1-4 means remove 1,2,3,4

        ex:  1-4 7-25-2 ~11-19-4

        yields

        1 2 3 4 7 9 13 17 21 23 25
        """)


def selected_selector(parser_args, defaults_to_all=False):
    """
    parser_args is the result of arg_parser.parse_args()
    """
    ranges = parser_args.ranges

    # our naming conventions
    selector = Selector()
    # nothing set on the command line : let's use $NODES
    if parser_args.all_nodes:
        selector.use_all_scope()
    # nothing specified at all - no range no --all-nodes
    if not ranges and not parser_args.all_nodes:
        if defaults_to_all:
            selector.use_all_scope()
        elif os.getenv('NODES'):
            for node in os.environ["NODES"].split():
                selector.add_range(node)
        else:
            print("no argument specified :"
                  " this requires you to set env. variable 'NODES'")
            exit(1)
    else:
        for range1 in ranges:
            selector.add_range(range1)

    return selector



# trap signals for a more condensed output
import signal
import functools

import asyncio

class MonitorLoop:

    def __init__(self, message):
        self.message = message

    def run(self, async_main, logger):

        loop = asyncio.get_event_loop()

        def exiting(signame):
            logger.info(f"rhubarbe-{self.message}: ¬ received signal {signame} - exiting")
            loop.stop()
            exit(1)

        for signame in ('SIGHUP', 'SIGQUIT', 'SIGINT', 'SIGTERM'):
            loop.add_signal_handler(getattr(signal, signame),
                                    functools.partial(exiting, signame))

        try:
            task = asyncio.ensure_future(async_main)
            loop.run_until_complete(task)
            return 0
        except KeyboardInterrupt:
            logger.info(f"rhubarbe-{self.message} : keyboard interrupt - exiting")
            task.cancel()
            loop.run_forever()
            task.exception()
            return 1
        except asyncio.TimeoutError:
            logger.info(f"rhubarbe-{self.message} : asyncio timeout expired")
            return 1
        finally:
            loop.close()



"""
The phones monitor cyclically checks for the status of all phones,
and reports it to the sidecar service

This simple tool works a bit like monitornodes, but on phones
for now it probes for airplane_mode, but does not probe for the state of
the wifi service because I could not get to work the magic sentences
like 'adb shell svc enable wifi' and similar that I have found on the web
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import asyncio

from apssh.sshproxy import SshProxy
from apssh.formatters import CaptureFormatter

from rhubarbe.config import Config
from rhubarbe.logger import monitor_logger as logger

from rhubarbe.inventoryphones import InventoryPhones
from rhubarbe.monitor.reconnectable import ReconnectableSidecar


class MonitorPhone:                                     # pylint: disable=r0902

    # id is what you get through adb devices
    def __init__(self, id,                       # pylint: disable=r0913, w0622
                 gw_host, gw_user, gw_key,
                 adb_bin, adb_id,
                 reconnectable, verbose, cycle=2):
        self.id = id                                    # pylint: disable=c0103
        # gateway is the macphone box in the middle
        self.gateway = SshProxy(
            hostname=gw_host,
            username=gw_user,
            keys=[gw_key],
            # setting verbose would end up on stdout
            formatter=CaptureFormatter(verbose=False)
        )
        self.adb_bin = adb_bin
        self.adb_id = adb_id
        self.reconnectable = reconnectable
        self.cycle = cycle
        self.info = {'id': self.id}
        self.verbose = verbose

    def __repr__(self):
        return ("monitored phone #{} via gw {}@{} with key {}"
                .format(self.id, self.gateway.username, self.gateway.hostname,
                        self.gateway.keys[0]))

    async def emit(self):
        await self.reconnectable.emit_info(self.info)

    async def probe(self):

        # connect or reconnect if needed
        if not self.gateway.is_connected():
            try:
                if self.verbose:
                    logger.info(f"{self}: connecting to gateway "
                                f"{self.gateway}")
                await self.gateway.connect_lazy()
                logger.info(f"Connected -> {self.gateway}")
            except Exception as exc:
                logger.error("Could not connect -> {} (exc={})"
                             .format(self.gateway, exc))
                self.info['airplane_mode'] = 'fail'
                await self.emit()

        if not self.gateway.is_connected():
            logger.error("Not connected to gateway - aborting")
            return

        try:
            if self.verbose:
                logger.info(f"{self}: retrieving airplane settings to gateway "
                            f"{self.gateway}")
            self.gateway.formatter.start_capture()
            retcod = await self.gateway.run(
                "{} shell \"settings get global airplane_mode_on\""
                .format(self.adb_bin))
            result = self.gateway.formatter.get_capture().strip()
            airplane_mode = 'fail' if retcod != 0 \
                else 'on' if result == '1' else 'off'
            if self.verbose:
                logger.info("probed phone {} : retcod={} result={} "
                            "-> airplane_mode = {}"
                            .format(self.adb_id, retcod, result,
                                    airplane_mode))
            self.info['airplane_mode'] = airplane_mode

        except Exception as exc:
            logger.error("Could not probe {} -> (e={})"
                         .format(self.adb_id, exc))
            self.info['airplane_mode'] = 'fail'
            # force ssh reconnect
            self.gateway.conn = None
        await self.emit()

    async def probe_forever(self):
        while True:
            await self.probe()
            await asyncio.sleep(self.cycle)


class MonitorPhones:                                     # pylint:disable=r0903
    def __init__(self, verbose, sidecar_url, cycle):
        self.verbose = verbose

        phone_specs = InventoryPhones().all_phones()

        self.reconnectable = ReconnectableSidecar(sidecar_url, 'phones')
        # xxx this is fragile
        # we rely on the fact that the items in the inventory
        # match the args of MonitorPhone's constructor
        self.phones = [
            MonitorPhone(reconnectable=self.reconnectable,
                         verbose=verbose,
                         cycle=cycle,
                         **spec)
            for spec in phone_specs]

    async def run_forever(self):
        await asyncio.gather(
            *[phone.probe_forever()
              for phone in self.phones],
            self.reconnectable.keep_connected())



"""
display class when using curses, like in
rhubarbe load --curses
"""

import curses

from rhubarbe.display import Display

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w0201,r1705


class DisplayCurses(Display):

    # extra room on top and on the left
    offsetl = 3
    offsetc = 25

    def start_hook(self):
        self.screen = curses.initscr()
        self.maxl, self.maxc = self.screen.getmaxyx()
        self.submaxc = self.maxc - self.offsetc
        if len(self.nodes) + self.offsetl + 2 >= self.maxl:
            self.screen.addstr(1, 1,
                               "WARNING: screen not high enough - "
                               "best effort..")
            self.screen.refresh()
            self.submaxl = self.maxl - self.offsetl
        else:
            self.submaxl = len(self.nodes) + 2
        self.usable_l = self.submaxl - 1
        self.screen.border()
        self.subwin = self.screen.subwin(self.submaxl, self.submaxc,
                                         self.offsetl, self.offsetc)
        self.subwin.border()

    # this is guaranteed to run once the event loop has returned
    # and all the messages have been displayed
    def epilogue(self):
        prompt = "Press any key to exit"
        if self.goodbye_message:
            prompt = self.goodbye_message + " " + prompt
        self.screen.addstr(self.maxl-1, 1, prompt)
        self.screen.refresh()
        self.screen.getch()
        curses.endwin()

    def repair(self):
        curses.endwin()

    def dispatch_hook(self, message, timestamp, duration):
        timemsg = f"{timestamp} {duration}"
        text = self.message_to_text(message)
        self.screen.addstr(1, 1, timemsg)
        self.screen.addstr(1, self.offsetc+1, self.pad(text))
        self.screen.refresh()

    def dispatch_ip_hook(self, _, node,            # pylint:disable=w0221,r0913
                         message, timestamp, duration):
        timemsg = f"{timestamp} {duration} {node.name}"
        text = self.message_to_text_ip(message, node, mention_node=False)
        text = str(text)
        line = (node.rank % self.usable_l) + 1
        self.screen.addstr(line+self.offsetl, 1, timemsg)
        self.subwin.addstr(line, 1, self.pad(text))
        self.screen.refresh()
        self.subwin.refresh()

    def dispatch_ip_percent_hook(self, _, node,    # pylint:disable=w0221,r0913
                                 message, timestamp, duration):
        # global area
        timemsg = f"{timestamp} {duration}"
        text = self.node_percent_bar(int((self.total_percent)/len(self.nodes)))
        self.screen.addstr(2, 1, timemsg)
        self.screen.addstr(2, self.offsetc+1, self.pad(text))
        # node area
        timemsg = f"{timestamp} {duration} {node.name}"
        barsize = self.node_percent_bar(node.percent)
        line = (node.rank % self.usable_l) + 1
        self.screen.addstr(line+self.offsetl, 1, timemsg)
        self.subwin.addstr(line, 1, barsize)
        self.screen.refresh()
        self.subwin.refresh()

    def node_percent_bar(self, percent):
        # 2 is for the 2 borders left and right; 4 is the size for '|10%'
        avail = self.submaxc - 2 - 4
        left = int(percent*avail/100)
        middle = avail - left
        right = f"|{percent:02}%" if percent != 100 else "||||"
        return left*'#' + middle*' ' + right

#    def pad(self, text):
#        return self._pad(text, self.maxc-2)
    def pad(self, text):
        return self._pad(text, self.submaxc-2)

    @staticmethod
    def _pad(text, width):
        if len(text) == width:
            return text
        elif len(text) > width:
            return text[:(width-3)] + '...'
        else:
            missing = width-len(text)
            return text + missing*' '



"""
Parse inventory json about phones
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# r0903 too few public methods
# pylint: disable=r0903

import json

from rhubarbe.singleton import Singleton
from rhubarbe.config import Config


class InventoryPhones(metaclass=Singleton):
    """
    A class for loading and storing the phones specifications
    typically from /etc/rhubarbe/inventory-phones.json
    """

    def __init__(self):
        conf = Config()
        try:
            with open(conf.value('testbed', 'inventory_phones_path')) as feed:
                self._phones = json.load(feed)
        except FileNotFoundError:
            self._phones = []

    def all_phones(self):
        """
        For now just return the json contents as is
        """
        return self._phones



"""
The nodes monitor cyclically checks for the status of all nodes,
and reports it to the sidecar service
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import time
import re
import asyncio

from rhubarbe.config import Config
from rhubarbe.node import Node
from rhubarbe.ssh import SshProxy
# use a dedicated logger for monitors
from rhubarbe.logger import monitor_logger as logger

# connect to sidecar
from rhubarbe.monitor.reconnectable import ReconnectableSidecar

# translate info into a single char for logging
def one_char_summary(info):
    if 'cmc_on_off' in info and info['cmc_on_off'] != 'on':
        return '.'
    if 'control_ping' in info and info['control_ping'] != 'on':
        return 'o'
    if 'control_ssh' in info and info['control_ssh'] != 'on':
        return '0'
    if 'os_release' in info and 'fedora' in info['os_release']:
        return 'F'
    if 'os_release' in info and 'ubuntu' in info['os_release']:
        return 'U'
    return '^'


class MonitorNode:
    """
    the logic for probing a node as part of monitoring
    formerly in fitsophia/website/monitor.py
    . first probes for the CMC status; if off: done
    . second probe for ssh; if on : done
    . third, checks for ping
    """

    def __init__(self, node, reconnectable,             # pylint: disable=r0913
                 report_wlan=False, verbose=False):
        # a rhubarbe.node.Node instance
        self.node = node
        self.report_wlan = report_wlan
        self.reconnectable = reconnectable
        self.verbose = verbose
        # current info - will be reported to sidecar
        self.info = {'id': node.id}
        # remember previous wlan measurement to compute rate
        self.history = {}

    def set_info(self, *overrides):
        """
        update self.info with all the dicts in overrides
        """
        for override in overrides:
            self.info.update(override)

    def zero_wlan_infos(self):
        """
        set wlan-related attributes to 0.
        """
        for k in self.info:
            if k.startswith('wlan'):
                self.info[k] = 0.

    async def report_info(self):
        """
        Send info to sidecar
        """
        await self.reconnectable.emit_info(self.info)

    async def set_info_and_report(self, *overrides):
        """
        Convenience to set info and immediately report
        """
        self.set_info(*overrides)
        await self.report_info()

    ubuntu_matcher = re.compile(r"DISTRIB_RELEASE=(?P<ubuntu_version>[0-9.]+)")
    fedora_matcher = re.compile(r"Fedora release (?P<fedora_version>\d+)")
    centos_matcher = re.compile(r"CentOS Linux release (?P<centos_version>\d+)")
    gnuradio_matcher = re.compile(
        r"\AGNURADIO:(?P<gnuradio_version>[0-9\.]+)\Z")
    uname_matcher = re.compile(r"\AUNAME:(?P<uname>.+)\Z")
    # 2016-05-28@08:20 - node fit38 - image oai-enb-base2 - by root
    rhubarbe_image_matcher = re.compile(
        r"\A/etc/rhubarbe-image:.* - image (?P<image_radical>[^ ]+) - by")
    rxtx_matcher = re.compile(
        r"==> /sys/class/net/wlan(?P<wlan_no>[0-9])/"
        r"statistics/(?P<rxtx>[rt]x)_bytes <==")
    number_matcher = re.compile(r"\A[0-9]+\Z")

    def parse_ssh_probe_output(self,      # pylint: disable=r0912, r0914, r0915
                               stdout, padding_dict):
        os_release = "other"
        gnuradio_release = "none"
        uname = ""
        rxtx_dict = {}
        rxtx_key = None
        image_radical = ""
        for line in stdout.split("\n"):
            match = self.ubuntu_matcher.match(line)
            if match:
                version = match.group('ubuntu_version')
                os_release = f"ubuntu-{version}"
                continue
            match = self.fedora_matcher.match(line)
            if match:
                version = match.group('fedora_version')
                os_release = f"fedora-{version}"
                continue
            match = self.centos_matcher.match(line)
            if match:
                version = match.group('centos_version')
                os_release = f"centos-{version}"
                continue
            match = self.gnuradio_matcher.match(line)
            if match:
                gnuradio_release = match.group('gnuradio_version')
                continue
            match = self.uname_matcher.match(line)
            if match:
                uname = match.group('uname')
                continue
            match = self.rhubarbe_image_matcher.match(line)
            if match:
                image_radical = match.group('image_radical')
                continue
            match = self.rxtx_matcher.match(line)
            if match:
                # use a tuple as the hash
                rxtx_key = (match.group('wlan_no'), match.group('rxtx'))
                continue
            match = self.number_matcher.match(line)
            if match and rxtx_key:
                rxtx_dict[rxtx_key] = int(line)
                continue
            rxtx_key = None

        # now that we have the counters we need to translate this into rate
        # for that purpose we use local clock;
        # small imprecision should not impact overall result
        now = time.time()
        wlan_info_dict = {}
        for rxtx_key, bytes in rxtx_dict.items():       # pylint: disable=w0622
            wlan_no, rxtx = rxtx_key
            # rather dirty hack for images that use wlan2 and wlan3
            # expose in wlan0 or wlan1 depending on parity of actual device
            try:
                wlan_no = int(wlan_no)
                wlan_no = wlan_no % 2
            except Exception:
                pass
            if self.verbose:
                logger.info(f"node={self.node} collected {bytes} "
                            f"for device wlan{wlan_no} in {rxtx}")
            # do we have something on this measurement ?
            if rxtx_key in self.history:
                previous_bytes, previous_time = self.history[rxtx_key]
                info_key = f"wlan_{wlan_no}_{rxtx}_rate"
                new_rate = 8.*(bytes - previous_bytes) / (now - previous_time)
                wlan_info_dict[info_key] = new_rate
                if self.verbose:
                    logger.info(f"node={id} computed {new_rate} bps for key {info_key} "
                                f"- bytes = {bytes}, pr = {previous_bytes}, "
                                f"now = {now}, pr = {previous_time}")
            # store this measurement for next run
            self.history[rxtx_key] = (bytes, now)
        # xxx would make sense to clean up history for measurements that
        # we were not able to collect at this cycle
        self.set_info({'os_release': os_release,
                       'gnuradio_release': gnuradio_release,
                       'uname': uname,
                       'image_radical': image_radical},
                      padding_dict, wlan_info_dict)

    async def probe(self,                 # pylint: disable=r0912, r0914, r0915
                    ping_timeout, ssh_timeout):
        """
        The logic for getting one node's info and send it to sidecar
        """
        if self.verbose:
            logger.info(f"entering pass1, info={self.info}")
        # pass1 : check for status
        padding_dict = {
            'control_ping': 'off',
            'control_ssh': 'off',
            # don't overwrite os_release though
        }
        # get USRP status no matter what - use "" if we receive None
        # to limit noise when the node is physically removed
        usrp_status = await self.node.get_usrpstatus() or 'fail'
        # replace usrpon and usrpoff with just on and off
        self.set_info({'usrp_on_off': usrp_status.replace('usrp', '')})
        # get CMC status
        status = await self.node.get_status()
        if status == "off":
            await self.set_info_and_report({'cmc_on_off': 'off'}, padding_dict)
            return
        elif status != "on":
            await self.set_info_and_report({'cmc_on_off': 'fail'}, padding_dict)
            return
        if self.verbose:
            logger.info(f"entering pass2, info={self.info}")
        # pass2 : CMC status is ON - let's try to ssh it
        self.set_info({'cmc_on_off': 'on'})
        padding_dict = {
            'control_ping': 'on',
            'control_ssh': 'on',
        }
        self.zero_wlan_infos()
        remote_commands = [
            "cat /etc/lsb-release /etc/redhat-release /etc/gnuradio-release "
            "2> /dev/null | grep -i release",
            "echo -n GNURADIO: ; gnuradio-config-info --version "
            "2> /dev/null || echo none",
            # this trick allows to have the filename on each output line
            "grep . /etc/rhubarbe-image /dev/null",
            "echo -n 'UNAME:' ; uname -r",
            ]
        if self.report_wlan:
            remote_commands.append(
                "head /sys/class/net/wlan?/statistics/[rt]x_bytes"
            )
        # reconnect each time
        async with SshProxy(self.node) as ssh:
            if self.verbose:
                logger.info(f"trying to ssh-connect to {self.node.control_hostname()} "
                            f"(timeout={ssh_timeout})")
            try:
                connected = await asyncio.wait_for(ssh.connect(),
                                                   timeout=ssh_timeout)
            except asyncio.TimeoutError:
                connected = False
                self.set_info({'control_ssh': 'off'})
            if self.verbose:
                logger.info(f"{self.node.control_hostname()} ssh-connected={connected}")
            if connected:
                try:
                    command = ";".join(remote_commands)
                    output = await asyncio.wait_for(ssh.run(command),
                                                    timeout=ssh_timeout)
                    # padding dict here sets control_ssh and control_ping to on
                    self.parse_ssh_probe_output(output, padding_dict)
                    # required as otherwise we leak openfiles
                    try:
                        await ssh.close()
                    except Exception:
                        logger.exception("monitornodes oops 1")
                except asyncio.TimeoutError:
                    self.set_info({'control_ssh': 'off'})
                except Exception:
                    logger.exception("monitornodes remote_command failed")
        if self.verbose:
            logger.info(f"{self.node.control_hostname()} ssh-based branch done "
                        f"ssh is deemed {self.info['control_ssh']}")

        # if we could ssh then we're done
        if self.info['control_ssh'] == 'on':
            await self.report_info()
            return

        if self.verbose:
            logger.info(f"entering pass3, info={self.info}")
        # pass3 : node is ON but could not ssh
        # check for ping
        # I don't know of an asyncio library to deal with icmp
        # so let's use asyncio.subprocess
        # xxx maybe a Ping class would be the way to go
        control = self.node.control_hostname()
        command = ["ping", "-c", "1", "-t", "1", control]
        try:
            subprocess = await asyncio.create_subprocess_exec(
                *command,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL)
            # failure occurs through timeout
            await asyncio.wait_for(subprocess.wait(),
                                   timeout=ping_timeout)
            await self.set_info_and_report({'control_ping': 'on'})
            return
        except asyncio.TimeoutError:
            await self.set_info_and_report({'control_ping': 'off'})
            return

    async def probe_forever(self, cycle, ping_timeout, ssh_timeout):
        """
        runs forever, wait <cycle> seconds between 2 runs of probe()
        """
        while True:
            try:
                await self.probe(ping_timeout, ssh_timeout)
            except Exception:
                logger.exception("monitornodes oops 2")
            await asyncio.sleep(cycle)


class MonitorNodes:                                     # pylint: disable=r0902

    def __init__(self, cmc_names, message_bus,          # pylint: disable=r0913
                 sidecar_url, cycle,
                 report_wlan=False, verbose=False):
        self.cycle = cycle
        self.report_wlan = report_wlan
        self.verbose = verbose

        # get miscell config
        self.ping_timeout = float(Config().value('networking', 'ping_timeout'))
        self.ssh_timeout = float(Config().value('networking', 'ssh_timeout'))
        self.log_period = float(Config().value('monitor', 'log_period'))

        # websockets
        self.reconnectable = \
            ReconnectableSidecar(sidecar_url, 'nodes')

        # the nodes part
        nodes = [Node(cmc_name, message_bus) for cmc_name in cmc_names]
        self.monitor_nodes = [
            MonitorNode(node=node, reconnectable=self.reconnectable,
                        report_wlan=self.report_wlan,
                        verbose=verbose)
            for node in nodes]

    async def log(self):
        previous = 0
        while True:
            line = "".join([one_char_summary(mnode.info)
                            for mnode in self.monitor_nodes])
            current = self.reconnectable.counter
            delta = f"+ {current-previous}"
            line += f" {current} emits ({delta})"
            previous = current
            logger.info(line)
            await asyncio.sleep(self.log_period)

    async def run_forever(self):
        logger.info(f"Starting nodes on {len(self.monitor_nodes)} nodes - "
                    f"report_wlan={self.report_wlan}")
        return asyncio.gather(
            *[monitor_node.probe_forever(self.cycle,
                                         ping_timeout=self.ping_timeout,
                                         ssh_timeout=self.ssh_timeout)
              for monitor_node in self.monitor_nodes],
            self.reconnectable.keep_connected(),
            self.log(),
        )
